@@ -59,11 +59,13 @@ from .screen import _CHAR_CACHE, Screen, WritePosition
 from .utils import explode_text_fragments
 
 if TYPE_CHECKING:
-    from typing_extensions import Protocol
+    from typing_extensions import Protocol, TypeGuard
+
+    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
 
-    NotImplementedOrNone = object
 
 __all__ = [
+    "AnyContainer",
     "Container",
     "HorizontalAlign",
     "VerticalAlign",
@@ -170,14 +172,14 @@ AnyContainer = Union[Container, "MagicContainer"]
 
 
 def _window_too_small() -> "Window":
-    " Create a `Window` that displays the 'Window too small' text. "
+    "Create a `Window` that displays the 'Window too small' text."
     return Window(
         FormattedTextControl(text=[("class:window-too-small", " Window too small... ")])
     )
 
 
 class VerticalAlign(Enum):
-    " Alignment for `HSplit`. "
+    "Alignment for `HSplit`."
     TOP = "TOP"
     CENTER = "CENTER"
     BOTTOM = "BOTTOM"
@@ -185,7 +187,7 @@ class VerticalAlign(Enum):
 
 
 class HorizontalAlign(Enum):
-    " Alignment for `VSplit`. "
+    "Alignment for `VSplit`."
     LEFT = "LEFT"
     CENTER = "CENTER"
     RIGHT = "RIGHT"
@@ -353,7 +355,8 @@ class HSplit(_Split):
                         style=self.padding_style,
                     )
                 )
-            result.pop()
+            if result:
+                result.pop()
 
             # Padding right.
             if self.align in (VerticalAlign.CENTER, VerticalAlign.TOP):
@@ -605,7 +608,8 @@ class VSplit(_Split):
                         style=self.padding_style,
                     )
                 )
-            result.pop()
+            if result:
+                result.pop()
 
             # Padding right.
             if self.align in (HorizontalAlign.CENTER, HorizontalAlign.LEFT):
@@ -748,7 +752,7 @@ class FloatContainer(Container):
                        floats=[
                            Float(xcursor=True,
                                 ycursor=True,
-                                layout=CompletionMenu(...))
+                                content=CompletionsMenu(...))
                        ])
 
     :param z_index: (int or None) When specified, this can be used to bring
@@ -858,7 +862,7 @@ class FloatContainer(Container):
         erase_bg: bool,
         z_index: Optional[int],
     ) -> None:
-        " Draw a single Float. "
+        "Draw a single Float."
         # When a menu_position was given, use this instead of the cursor
         # position. (These cursor positions are absolute, translate again
         # relative to the write_position.)
@@ -1064,7 +1068,7 @@ class Float:
         allow_cover_cursor: bool = False,
         z_index: int = 1,
         transparent: bool = False,
-    ):
+    ) -> None:
 
         assert z_index >= 1
 
@@ -1105,7 +1109,7 @@ class Float:
 
 class WindowRenderInfo:
     """
-    Render information, for the last render time of this control.
+    Render information for the last render time of this control.
     It stores mapping information between the input buffers (in case of a
     :class:`~prompt_toolkit.layout.controls.BufferControl`) and the actual
     render position on the output screen.
@@ -1571,8 +1575,8 @@ class Window(Container):
         """
 
         def preferred_content_width() -> Optional[int]:
-            """ Content width: is only calculated if no exact width for the
-            window was given. """
+            """Content width: is only calculated if no exact width for the
+            window was given."""
             if self.ignore_content_width():
                 return None
 
@@ -1602,8 +1606,8 @@ class Window(Container):
         """
 
         def preferred_content_height() -> Optional[int]:
-            """ Content height: is only calculated if no exact height for the
-            window was given. """
+            """Content height: is only calculated if no exact height for the
+            window was given."""
             if self.ignore_content_height():
                 return None
 
@@ -1684,7 +1688,7 @@ class Window(Container):
         return self._ui_content_cache.get(key, get_content)
 
     def _get_digraph_char(self) -> Optional[str]:
-        " Return `False`, or the Digraph symbol to be used. "
+        "Return `False`, or the Digraph symbol to be used."
         app = get_app()
         if app.quoted_insert:
             return "^"
@@ -1707,6 +1711,32 @@ class Window(Container):
         Write window to screen. This renders the user control, the margins and
         copies everything over to the absolute position at the given screen.
         """
+        # If dont_extend_width/height was given. Then reduce width/height in
+        # WritePosition if the parent wanted us to paint in a bigger area.
+        # (This happens if this window is bundled with another window in a
+        # HSplit/VSplit, but with different size requirements.)
+        write_position = WritePosition(
+            xpos=write_position.xpos,
+            ypos=write_position.ypos,
+            width=write_position.width,
+            height=write_position.height,
+        )
+
+        if self.dont_extend_width():
+            write_position.width = min(
+                write_position.width,
+                self.preferred_width(write_position.width).preferred,
+            )
+
+        if self.dont_extend_height():
+            write_position.height = min(
+                write_position.height,
+                self.preferred_height(
+                    write_position.width, write_position.height
+                ).preferred,
+            )
+
+        # Draw
         z_index = z_index if self.z_index is None else self.z_index
 
         draw_func = partial(
@@ -1800,13 +1830,16 @@ class Window(Container):
         self.render_info = render_info
 
         # Set mouse handlers.
-        def mouse_handler(mouse_event: MouseEvent) -> None:
-            """ Wrapper around the mouse_handler of the `UIControl` that turns
-            screen coordinates into line coordinates. """
+        def mouse_handler(mouse_event: MouseEvent) -> "NotImplementedOrNone":
+            """
+            Wrapper around the mouse_handler of the `UIControl` that turns
+            screen coordinates into line coordinates.
+            Returns `NotImplemented` if no UI invalidation should be done.
+            """
             # Don't handle mouse events outside of the current modal part of
             # the UI.
             if self not in get_app().layout.walk_through_modal_area():
-                return
+                return NotImplemented
 
             # Find row/col position first.
             yx_to_rowcol = {v: k for k, v in rowcol_to_yx.items()}
@@ -1832,6 +1865,8 @@ class Window(Container):
                         MouseEvent(
                             position=Point(x=col, y=row),
                             event_type=mouse_event.event_type,
+                            button=mouse_event.button,
+                            modifiers=mouse_event.modifiers,
                         )
                     )
                     break
@@ -1842,13 +1877,18 @@ class Window(Container):
                 # Report (0,0) instead.)
                 result = self.content.mouse_handler(
                     MouseEvent(
-                        position=Point(x=0, y=0), event_type=mouse_event.event_type
+                        position=Point(x=0, y=0),
+                        event_type=mouse_event.event_type,
+                        button=mouse_event.button,
+                        modifiers=mouse_event.modifiers,
                     )
                 )
 
             # If it returns NotImplemented, handle it here.
             if result == NotImplemented:
-                self._mouse_handler(mouse_event)
+                result = self._mouse_handler(mouse_event)
+
+            return result
 
         mouse_handlers.set_mouse_handler_for_range(
             x_min=write_position.xpos + sum(left_margin_widths),
@@ -1862,7 +1902,7 @@ class Window(Container):
         move_x = 0
 
         def render_margin(m: Margin, width: int) -> UIContent:
-            " Render margin. Return `Screen`. "
+            "Render margin. Return `Screen`."
             # Retrieve margin fragments.
             fragments = m.create_margin(render_info, width, write_position.height)
 
@@ -1894,8 +1934,9 @@ class Window(Container):
         # Apply 'self.style'
         self._apply_style(screen, write_position, parent_style)
 
-        # Tell the screen that this user control has been painted.
-        screen.visible_windows.append(self)
+        # Tell the screen that this user control has been painted at this
+        # position.
+        screen.visible_windows_to_write_positions[self] = write_position
 
     def _copy_body(
         self,
@@ -2018,7 +2059,7 @@ class Window(Container):
                             return x, y  # Break out of all for loops.
 
                     # Set character in screen and shift 'x'.
-                    if x >= 0 and y >= 0 and x < write_position.width:
+                    if x >= 0 and y >= 0 and x < width:
                         new_buffer_row[x + xpos] = char
 
                         # When we print a multi width character, make sure
@@ -2080,7 +2121,7 @@ class Window(Container):
         copy()
 
         def cursor_pos_to_screen_pos(row: int, col: int) -> Point:
-            " Translate row/col from UIContent to real Screen coordinates. "
+            "Translate row/col from UIContent to real Screen coordinates."
             try:
                 y, x = rowcol_to_yx[row, col]
             except KeyError:
@@ -2444,7 +2485,7 @@ class Window(Container):
             window_size: int,
             content_size: int,
         ) -> int:
-            " Scrolling algorithm. Used for both horizontal and vertical scrolling. "
+            "Scrolling algorithm. Used for both horizontal and vertical scrolling."
             # Calculate the scroll offset to apply.
             # This can obviously never be more than have the screen size. Also, when the
             # cursor appears at the top or bottom, we don't apply the offset.
@@ -2517,18 +2558,25 @@ class Window(Container):
             ),
         )
 
-    def _mouse_handler(self, mouse_event: MouseEvent) -> None:
+    def _mouse_handler(self, mouse_event: MouseEvent) -> "NotImplementedOrNone":
         """
         Mouse handler. Called when the UI control doesn't handle this
         particular event.
+
+        Return `NotImplemented` if nothing was done as a consequence of this
+        key binding (no UI invalidate required in that case).
         """
         if mouse_event.event_type == MouseEventType.SCROLL_DOWN:
             self._scroll_down()
+            return None
         elif mouse_event.event_type == MouseEventType.SCROLL_UP:
             self._scroll_up()
+            return None
+
+        return NotImplemented
 
     def _scroll_down(self) -> None:
-        " Scroll window down. "
+        "Scroll window down."
         info = self.render_info
 
         if info is None:
@@ -2541,7 +2589,7 @@ class Window(Container):
             self.vertical_scroll += 1
 
     def _scroll_up(self) -> None:
-        " Scroll window up. "
+        "Scroll window up."
         info = self.render_info
 
         if info is None:
@@ -2616,7 +2664,7 @@ class ConditionalContainer(Container):
 
 class DynamicContainer(Container):
     """
-    Container class that can dynamically returns any Container.
+    Container class that dynamically returns any Container.
 
     :param get_container: Callable that returns a :class:`.Container` instance
         or any widget with a ``__pt_container__`` method.
@@ -2644,8 +2692,18 @@ class DynamicContainer(Container):
     def preferred_height(self, width: int, max_available_height: int) -> Dimension:
         return self._get_container().preferred_height(width, max_available_height)
 
-    def write_to_screen(self, *a, **kw):
-        self._get_container().write_to_screen(*a, **kw)
+    def write_to_screen(
+        self,
+        screen: Screen,
+        mouse_handlers: MouseHandlers,
+        write_position: WritePosition,
+        parent_style: str,
+        erase_bg: bool,
+        z_index: Optional[int],
+    ) -> None:
+        self._get_container().write_to_screen(
+            screen, mouse_handlers, write_position, parent_style, erase_bg, z_index
+        )
 
     def is_modal(self) -> bool:
         return False
@@ -2670,9 +2728,9 @@ def to_container(container: AnyContainer) -> Container:
     if isinstance(container, Container):
         return container
     elif hasattr(container, "__pt_container__"):
-        return to_container(cast("MagicContainer", container).__pt_container__())
+        return to_container(container.__pt_container__())
     else:
-        raise ValueError("Not a container object.")
+        raise ValueError("Not a container object: %r" % (container,))
 
 
 def to_window(container: AnyContainer) -> Window:
@@ -2687,7 +2745,7 @@ def to_window(container: AnyContainer) -> Window:
         raise ValueError("Not a Window object: %r." % (container,))
 
 
-def is_container(value: object) -> bool:
+def is_container(value: object) -> "TypeGuard[AnyContainer]":
     """
     Checks whether the given value is a container object
     (for use in assert statements).
